// Module echtzeit.cpp
// Copyright Dr.E.Huckert 12-2010
// Demonstrates:
//   - Creation, termination and use of two IO parallel threads
//   - Use of lock mechanisms (mutexes)
//   - Use of message queues between threads

// Compile (Digital Mars C++ compiler):
//    dmc -mn -D_WINSOCKAPI_ echtzeit.cpp mutex.cpp hu_q.cpp 
//         threadbase.cpp kbthread.cpp udpthread.cpp 
//         hu_trace.cpp wsock32.lib

#include <stdio.h>
#include "kbthread.h"
#include "udpthread.h"
#include "hu_q.h"
#include "mutex.h"
#include "hu_trace.h"

// global variables
HU_TRACE *pTrace     = new HU_TRACE("REALTIME", "localhost", 9514);
Mutex *pFlagMutex    = new Mutex("FLAGS");
int main_go_on       = 1;
Mutex *pMainM        = new Mutex("MAINQ");  // mutex for the main queue
HU_Q *pMainQ         = new HU_Q(100);       // main queue
//
Mutex *pKbInpM       = new Mutex("KBINP");  // mutex for the keyboard inp. queue
HU_Q *pKbInpQ        = new HU_Q(10);        // main keyboard inp. queue
KbThread *pKbThread = NULL;
//
Mutex *pUdpInpM       = new Mutex("UDPINP");  // mutex for the keyboard inp. queue
HU_Q *pUdpInpQ       = new HU_Q(10);        // UDP inp. queue
UDPThread *pUdpThread = NULL;
int udpPort           = 17777;


// -----------------------------------------------------------------------
// CTRL-C Handler - is invoked when CTL-C is pressed  
// We set here the flag for the main thread to 0 so that
// the main event loop is finished
BOOL WINAPI ctrlcHandler(DWORD event_type)
{
  if (::pTrace != NULL)
    ::pTrace->trace("ctrlcHandler() activated");
  ::pFlagMutex->setlock();
  ::main_go_on     = 0;
  ::pFlagMutex->unlock();
  return TRUE;  
}   // end  ctrlcHandler()

// ------------------------------------------------------------
int evaluateCommandLine(int argc, char *argv[])
{
  int ret = 0;
  if (::pTrace != NULL)
    ::pTrace->trace("evaluateCommandLine()started");
  // ... nothing to do here at the moment
  return ret;
}   // end evaluateCommandLine()

// ------------------------------------------------------------
// Start all IO threads
// We start two threads here:
//   A keyboard input thread
//   A UDP input thread
// Note: A third thread - the main thread - is already started
int startIOThreads()
{
  int ret = 0;
  pKbThread = new KbThread(pKbInpQ, pKbInpM, pMainQ, pMainM);
  pKbThread->setLogging(::pTrace);
  pKbThread->createAndRun();
  if (::pTrace != NULL)
    ::pTrace->trace("startIOThreadS() thread 1 started");
  ret++;
  //
  pUdpThread = new UDPThread(::udpPort, pUdpInpQ, pUdpInpM, pMainQ, pMainM);
  pUdpThread->setLogging(::pTrace);
  pUdpThread->createAndRun();
  if (::pTrace != NULL)
    ::pTrace->trace("startIOThreadS() thread 2 started");
  ret++;
  return ret;
}   // end startIOThreads()

// ------------------------------------------------------------
// stop all IO threads
// We put "exit" commands into the input queue for the threads.
// These commands will be read by the threads - the thread will then
// stop themselves
int stopIOThreads()
{
  int ret = 0;
  ::pKbInpM->setlock();
  ::pKbInpQ->add("exit");
  ::pKbInpM->unlock();
  if (::pTrace != NULL)
    ::pTrace->trace("stopIOThreadS() thread 1 stopped");
  ret++;
  //
  ::pUdpInpM->setlock();
  ::pUdpInpQ->add("exit");
  ::pUdpInpM->unlock();
  if (::pTrace != NULL)
    ::pTrace->trace("stopIOThreadS() thread 2 stopped");
  ret++;
  return ret;
}   // end stopIOThreads()

// ------------------------------------------------------------
int main(int argc, char *argv[])
{
  int cRet;
  char *pCmd;//
  // evalute the command line
  cRet = evaluateCommandLine(argc, argv);
  //
  // start the IO threads
  cRet = startIOThreads();
  if (::pTrace != NULL)
    ::pTrace->trace("IO threads started result=", cRet);
  //
  // We are  here the main thread.
  // We start the main event loop.
  // Note: main_go_on is set to zero in the CTL-C handler
  SetConsoleCtrlHandler(::ctrlcHandler, TRUE);
  //
  while (1)
  {
    // go_on flag still set?
    ::pFlagMutex->setlock();
    if (! ::main_go_on)
    {
      ::pFlagMutex->unlock();
      break;
    }
    ::pFlagMutex->unlock();
    //
    // Inspect the main queue
    ::pMainM->setlock();
    cRet = pMainQ->getNoEntries();
    if (cRet > 0)
      // read the message in the main queue
      pCmd = pMainQ->get();
    else
      pCmd = NULL;
    ::pMainM->unlock();
    //
    if (pCmd != NULL)
    {
      // evaluate the message from the main queue
      if (::pTrace != NULL)
       ::pTrace->trace("main thread: message=", pCmd);
      delete pCmd;
    }
    //
    ::Sleep(100);
  }   // end while (1)
  //
  // stop all IO threads
  cRet = stopIOThreads();
  if (::pTrace != NULL)
    ::pTrace->trace("IO threads stopped result=", cRet);
  //
  // We sleep here in order to ensure that all threads are terminated
  ::Sleep(250);
  //
  if (pFlagMutex != NULL)
    delete pFlagMutex;
  if (pMainM != NULL)
    delete pMainM;
  if (pMainQ != NULL)
    delete pMainQ;
  if (::pTrace != NULL)
  {
    delete ::pTrace;
    ::pTrace = NULL;
  }
  ::exit(0);
}   // end main()