// module hu_q.cpp
// A minimalistic class for (non synchronized) FIFO queues implemented 
// via a singly chained list.
// Private Copyright Dr.E.Huckert 03-2009 

// Tested under Windows/Digital Mars C++ and
//        under Linux/g++
// Compile under Windows: dmc -mn -DWIN32 -DSTANDALONE hu_q.cpp
// Compile under Linux: g++ -DSTANDALONE hu_q.cpp

// Versions:
// 01.06.2010 made compatible for binary queue contents

#include "hu_q.h"

// ----------------------------------------------------------------
// Constructor for HU_QE
// For non-binary contents
HU_QE::HU_QE(const char *str)
{
  int  ll;
  ll = ::strlen(str);
  this->pContent = new char[ll+1];
  if (this->pContent == NULL)
  {
    // Exception hier werfen    
    return;
  }
  ::strcpy(this->pContent, str);
  this->lContent = ll;
  this->pNext = NULL;
}   // end HU_QE::HU_QE() 

// ----------------------------------------------------------------
// Constructor for HU_QE
// For binary contents
HU_QE::HU_QE(const char *str, int lStr)
{
  this->pContent = new char[lStr];
  if (this->pContent == NULL)
  {
    // Exception hier werfen    
    return;
  }
  ::memcpy(this->pContent, str, lStr);
  this->lContent = lStr;
  this->pNext = NULL;
}   // end HU_QE::HU_QE() 

// ----------------------------------------------------------------
// Destructor
HU_QE::~HU_QE()
{
  if (this->pContent != NULL)
    delete this->pContent;
  this->pContent = NULL;
}

// ----------------------------------------------------------------
// ----------------------------------------------------------------
// Constructor for HU_Q
HU_Q::HU_Q()
{
  this->pStart      = NULL;
  this->noEntries   = 0;
  this->maxEntries  = HUQ_MAXENTRIES;
}   // end HU_Q::HU_Q()

// ----------------------------------------------------------------
// Constructor for HU_Q
HU_Q::HU_Q(unsigned max_entries)
{
  this->pStart      = NULL;
  this->noEntries   = 0;
  this->maxEntries  = max_entries;
}   // end HU_Q::HU_Q

// ----------------------------------------------------------------
// Destructor
// frees the memory for all the list elements
HU_Q::~HU_Q()
{
  this->clear();
}   // end HU_Q::~HU_Q()

// ----------------------------------------------------------------
HU_QE *HU_Q::getPrevious(HU_QE *pe)
{
  HU_QE *pv = NULL;
  HU_QE *p = this->pStart;
  while (p != NULL)
  {
    if (p->getNext() == pe)
    {
      pv = p;
      break;
    }
    p = p->getNext();
  }
  return pv;
}   // end HU_Q::getPrevious()

// ----------------------------------------------------------------
// Remove the last element from the queue
void HU_Q::removeLast(HU_QE *pe)
{
  if (pe == NULL)
    return;
  HU_QE *pv = this->getPrevious(pe);
  if (pv != NULL)
    pv->setNext(NULL);
  if (pe == this->pStart)
    this->pStart = NULL;
  delete pe;
}

// ----------------------------------------------------------------
// Return the content of the last element from the FIFO queue
// This element is then removed.
// The returned char pointer must be freed externally!
// May return NULL!
// For non-binary contents
char * HU_Q::get()
{
  HU_QE *pLast;
  HU_QE *p = this->getStart();
  char *pc = NULL;
  while (p != NULL)
  {
    pLast = p;
    p = p->getNext();
  }
  if (pLast != NULL)
  {
    int ll = ::strlen(pLast->getContent()) + 1;
    pc = new char[ll];
    ::strcpy(pc, pLast->getContent());
    this->removeLast(pLast);
    this->noEntries--;
  }
  return pc;   
}   // end HU_Q::get()

// ----------------------------------------------------------------
// Return the content of the last element from the FIFO queue
// This element is then removed.
// The returned char pointer must be freed externally!
// May return NULL!
// Returns the length of the extracted content in return parameter pLength
// For binary contents
char * HU_Q::get(int *pLength)
{
  HU_QE *pLast;
  HU_QE *p = this->getStart();
  char *pc = NULL;
  *pLength = 0;
  while (p != NULL)
  {
    pLast = p;
    p = p->getNext();
  }
  if (pLast != NULL)
  {
    pc = new char[pLast->getContentLength()];
    ::memcpy(pc, pLast->getContent(), pLast->getContentLength());
    this->removeLast(pLast);
    this->noEntries--;
    *pLength = pLast->getContentLength();
  }
  return pc;   
}   // end HU_Q::get()

// ----------------------------------------------------------------
// Remove all entries from the queue
void HU_Q::clear()
{
  HU_QE *pn;
  HU_QE *p = this->getStart();
  while (p != NULL)
  {
    pn = p->getNext();
    delete p;
    p = pn;
  }
  this->noEntries = 0;
  this->pStart = NULL;
}   // end HU_Q::clear()

// ----------------------------------------------------------------
// get a pointer to the start of the list
HU_QE *HU_Q::getStart()
{
  return this->pStart;
}  // end HU_Q::get_start

// ----------------------------------------------------------------
// get a pointer to the end of the list 
HU_QE *HU_Q::getEnd()
{
  HU_QE *pLast = NULL;
  HU_QE *p = this->getStart();
  while (p != NULL)
  {
    pLast = p;
    p = p->getNext();
  }
  return pLast;
}  // end HU_Q::get_end

// ----------------------------------------------------------------
// Add an entry to the start of the FIFO chain
// Returns 0 if OK, < 0 if error
// This is for non-binary contents
int HU_Q::add(const char *ps)
{
  int ret = 0;
  HU_QE *pe;
  if (ps == NULL)
    return -1;
  if (this->noEntries >= this->maxEntries)
    return -2;
  pe = new HU_QE(ps);
  pe->setNext(this->pStart);
  this->pStart = pe;
  this->noEntries++;
  return ret;
}   // end HU_Q::add()

// ----------------------------------------------------------------
// Add an entry to the start of the FIFO chain
// Returns 0 if OK, < 0 if error
// This is for binary contents
int HU_Q::add(const char *ps,
              int lps)
{
  int ret = 0;
  HU_QE *pe;
  if (ps == NULL)
    return -1;
  if (this->noEntries >= this->maxEntries)
    return -2;
  pe = new HU_QE(ps, lps);
  pe->setNext(this->pStart);
  this->pStart = pe;
  this->noEntries++;
  return ret;
}   // end HU_Q::add()

// ----------------------------------------------------------------
// get the number of entries in the queue 
unsigned HU_Q::getNoEntries()
{
  return this->noEntries;
}  // end HU_Q::getNoEntries()

// ----------------------------------------------------------------
#ifdef STANDALONE
// ----------------------------------------------------------------
#ifdef WIN32
// for Digital Mars compiler
#include <iostream.h>
#else
#include <iostream>
using namespace std;
#endif

int main(int argc, char *argv[])
{
  cout << "testing classes HU_Q and HU_QE" << endl;
  HU_Q *pList = new HU_Q(5);
  // basic list construction
  cout << "Inserting in order 1..5" << endl;
  int ret;
  ret = pList->add("1");
  ret = pList->add("22");
  ret = pList->add("333");
  ret = pList->add("4444");
  ret = pList->add("55555");
  ret = pList->add("666666"); // results in an error
  cout << "result last add operation=" << ret << endl;
  //
  // abklappern - nicht loeschen
  cout << "walking the chain from start to end" << endl;
  HU_QE *p = pList->getStart();
  while (p != NULL)
  {
    cout << p->getContent() << endl;
    p = p->getNext();
  }
  cout << "no entries=" << pList->getNoEntries() << endl;
  //
  // Inhalte holen und loeschen
  cout << "emptying the chain and getting the content" << endl;
  char *pc;
  while (pList->getStart() != NULL)
  {
    pc = pList->get();
    cout << pc << endl;
    delete pc;
  }
  cout << "no entries=" << pList->getNoEntries() << endl;
  delete pList;  
}
#endif
