// module UDPThread.cpp

#include "udpthread.h"

static WSADATA WSAdata;

// ---------------------------------------------------------
// Constructor
// Creates a thread instance - does not start the thread!
// Opens and binds a Datagramm (UDP) socket
// Note: if errors occur method getNoErrors() returns >0!
UDPThread::UDPThread(int udpPort,
                     HU_Q *piq, Mutex *pim,
                     HU_Q *poq, Mutex *pom) :
                     ThreadBase(piq, pim, poq, pom)
{
  this->threadId = 0;
  this->pTrace = NULL;
  this->errors = 0;
  this->port = udpPort;
  const int y = 1;
  // Windows needs this (no equivalent in Unix/Linux)
  if (WSAStartup(0x0101, &(::WSAdata)) != 0)
  {
    this->errors++;
    return;
  }
  // create and bind the socket
  this->sock = ::socket(AF_INET, SOCK_DGRAM, 0);
  if (this->sock < 0)
  {
    this->errors++;
    return;
  }
  ::setsockopt(this->sock, SOL_SOCKET,
              SO_REUSEADDR, (char *)(&y), sizeof(int));
  // Note: we bind here only for receiving - the send side would be 
  // a little bit more complex
  ::memset(&(this->udpSender), 0, sizeof(this->udpSender));
  this->udpSender.sin_family      = AF_INET;
  this->udpSender.sin_port        = ::htons(this->port);
  if (::bind(this->sock, 
            (struct sockaddr*)&(this->udpSender), sizeof(this->udpSender)) < 0)
    this->errors++;
}

// ---------------------------------------------------------
UDPThread::~UDPThread()
{
  ::closesocket(this->sock);
}

// ----------------------------------------------------------
void UDPThread::setLogging(HU_TRACE *pt)
{
  this->pTrace = pt;
}

// ----------------------------------------------------------------
// Receive from the socket
// Returns: < 0 upon error
//          = 0: nothing received
//          > 0: something received
int UDPThread::receive(char *data, size_t size)
{
   int len, ret;
   ::memset(data, 0, size);
   len = sizeof (this->udpSender);
   ret = ::recvfrom(this->sock, data, size, 0,
                     (struct sockaddr *)&(this->udpSender), &len);
   if (ret == SOCKET_ERROR)
     ret = -1;
   else
     ret = len;
   return ret;
}   // end UDPThread::receive()

// ----------------------------------------------------
// thread worker routine
//void *threadWorker_udp(void *threadid)
DWORD WINAPI threadWorker_udp(void *param)
{
  int ret   = 0;
  int cRet;
  char msg[256];
  UDPThread *pk;
  //
  // get a pointer the the UDPThread instance
  pk = (UDPThread *)param;
  //
  while (1)
  {
    ::Sleep(10);
    // 
    // are there commands in out input queue?
    msg[0] = 0;
    (pk->getInpMutex())->setlock();
    cRet = (pk->getInpQueue())->getNoEntries();
    if (cRet > 0)
    {
      // read the message in the input queue
      char *pCmd = (pk->getInpQueue())->get();
      ::strcpy(msg, pCmd);
      delete pCmd;
    }
    (pk->getInpMutex())->unlock();
    if (::strlen(msg) > 0)
    {
      if (::strncmp(msg, "exit", 4) == 0)
        // "exit" command detected - leave this loop
        // leave this thread!
        break;
    }
    //
    // receive via UDP
    cRet = pk->receive(msg, sizeof(msg));
    if (cRet > 0)
    {
      if (pk->pTrace != NULL)
        pk->pTrace->trace("threadWorker_udp(): RX=", msg);
      // 
      // put the message just read into the queue for the main thread
      (pk->getOutMutex())->setlock();
      (pk->getOutQueue())->add(msg);
      (pk->getOutMutex())->unlock();
    }
  }   // end while (1)
  //
  if (pk->pTrace != NULL)
    pk->pTrace->trace("threadWorker_udp(): exiting");
  return (DWORD)ret;
}   // end threadWorker_udp()

// ---------------------------------------------------------
// create and start a thread
// Returns 0 if OK
int UDPThread::createAndRun()
{
  int ret = 0;
  HANDLE hThread;
  // Start a thread that will read the DB queue and
  // execute the DB requests    
  hThread = CreateThread(NULL, 
                         (DWORD)(16L * 1024L),   // stack size in bytes      
                         threadWorker_udp,        // thread worker routine
                         this,                   // argument to be passed
                         0, 
                         &(this->threadId));
  if (this->pTrace != NULL)
      this->pTrace->trace("UDPThread::createAndRun() ID=", (long)threadId);
  return ret;
}   // end UDPThread::create()